-- yadafrac : yet another fractal exploration program, as a GtkAda training
--
-- Written by Thierry Bernier, see http://thiberlog.free.fr/
-- this source code is released to the public domain.
--
-- $Id: gadtal.adb,v 1.5 2005/02/25 18:06:30 Grands Exp $


with Ada.Text_IO;         use Ada.Text_IO;

with Glib;                use Glib;
with Gdk.Drawable;        use Gdk.Drawable;
with Gdk.Pixmap;          use Gdk.Pixmap;
with Gdk.Event;           use Gdk.Event;
with Gdk.GC;              use Gdk.GC;
with Gtk.Enums;           use Gtk.Enums;
with Gtk.Handlers;        use Gtk.Handlers;
with Gtk.Arguments;       use Gtk.Arguments;
with Gtk.Box;             use Gtk.Box;
with Gtk.Button;          use Gtk.Button;
with Gtk.Spin_Button;     use Gtk.Spin_Button;
with Gtk.Main;            use Gtk.Main;

with Number;

package body Gadtal is

  use type Number.Real;
  use type Surface.Color_Type;
  use type Gdk_Drawable, Gdk_GC;

  Base_Rect : constant Number.Rect := (Min => (Re => -2.3, Im => -1.6),
                                       Max => (Re =>  1.2, Im =>  1.6));


  function Mandelbrot (C : Number.Complex; D : Value_Type) return Value_Type is
    X, Y   : Number.Real := 0.0;
    X2, Y2 : Number.Real;
    V      : Value_Type := D;
  begin
    while V > 0 loop
      X2 := X * X;
      Y2 := Y * Y;
      exit when X2 + Y2 > 4.0;
      Y := 2.0*X*Y + C.Im;
      X := X2 - Y2 + C.Re;
      V := V - 1;
    end loop;
    return V;
  end Mandelbrot;


  --
  procedure Evaluate (Gad  : access Gadtal_Record'Class;
                      Fmap : Fmap_Type;
                      Rect : Number.Rect)
  is
    pragma Debug (Put_Line ("Evaluate"));

    S : Surface.Surface_Type;

    -- Taille minimale pour decomposer un rectangle
    Min_Size : constant Gint := 8;

    -- Un sous-rectangle a evaluer
    type Pixrect_Type is record
      Xinf, Xsup, Yinf, Ysup : Gint;
    end record;

    -- Pile des rectangles a evaluer
    subtype Stack_Index_Type is Natural range 0 .. 1000;
    Stack     : array (Stack_Index_Type) of Pixrect_Type;
    Stack_Top : Stack_Index_Type := 0;

    -- Tampon des conversions pixel -> complexe
    type Rmap is array (Gint range <>) of Number.Real;
    Xmap : Rmap (0 .. Gad.Width-1);
    Ymap : Rmap (0 .. Gad.Height-1);

    -- Profondeur maxi
    Depth : constant Value_Type := Value_Type (Get_Value (Gad.Adj_Depth));
    CtoD  : constant Number.Real :=   Number.Real (Surface.Color_Type'Last)
                                    / Number.Real (Depth);

    -- Construire les tableaux de conversion pixel -> complexe
    procedure Build (M  : out Rmap; Rmin, Rmax : in Number.Real)
    is
      pragma Assert (M'First = 0);
      K : constant Number.Real := (Rmax - Rmin) / Number.Real (M'Length);
    begin
      for I in M'Range loop M(I) := Rmin + K * Number.Real(I); end loop;
    end;

    -- Evaluer un point
    function Point (X, Y : Gint) return Value_Type is
      V : constant Value_Type
          := Fmap ((Re => Xmap(X), Im => Ymap(Y)), Depth);
      C : constant Surface.Color_Type
          := Surface.Color_Type (Number.Real(V) * CtoD);
    begin
      Surface.Draw_Point (S, X, Y, C);
      return C;
    end Point;

    -- Evaluer l'interieur d'un rectangle
    procedure Interior (R : in Pixrect_Type) is
      Dummy : Value_Type;
      pragma Warnings (Off, Dummy);	-- Assigned but never read
    begin
      for I in R.Xinf+1 .. R.Xsup-1 loop for J in R.Yinf+1 .. R.Ysup-1 loop
        Dummy := Point (I, J);
      end loop; end loop;
    end;

    -- Remplir un rectangle
    procedure Fill (R : in Pixrect_Type; V : in Value_Type) is
    begin
      Surface.Fill_Rect (S    => S, Xmin => R.Xinf + 1, Xmax => R.Xsup - 1,
                                    Ymin => R.Yinf + 1, Ymax => R.Ysup - 1,
                         C    => V);
    end;

    -- Evaluer le contour d'un rectangle
    procedure Contour (R       : in  Pixrect_Type;
                       Uniform : out Boolean;
                       V       : out Value_Type) is
    begin
      Uniform := True;
      V := Point (R.Xinf, R.Yinf);
      if Point (R.Xinf, R.Ysup) /= V then Uniform := False; end if;
      for I in R.Xinf+1 .. R.Xsup loop
        if Point (I, R.Yinf) /= V or Point (I, R.Ysup) /= V then
          Uniform := False;
        end if;
      end loop;
      for J in R.Yinf+1 .. R.Ysup-1 loop
        if Point (R.Xinf, J) /= V or Point (R.Xsup, J) /= V then
          Uniform := False;
        end if;
      end loop;
    end;

  begin
--  Surface.Reset (S, Get_Window (Gad.Area));
    Surface.Reset (S, Gad.Store);

    -- Initialiser les tables
    Build (Xmap, Number.Real_CT.Re (Rect.Min), Number.Real_CT.Re (Rect.Max));
    Build (Ymap, Number.Real_CT.Im (Rect.Min), Number.Real_CT.Im (Rect.Max));

    -- Liste initiale de rectangle : le tout
    Stack(1) := (Xinf => 0, Xsup => Surface.Width(S) - 1,
                 Yinf => 0, Ysup => Surface.Height(S) - 1);
    Stack_Top := 1;

    while Stack_Top > 0 and then not Gad.Halt loop
      declare
        Job  : constant Pixrect_Type := Stack(Stack_Top);
        V    : Value_Type;
        Uni  : Boolean;			-- "Bords uniformes"
      begin
        Contour (Job, Uni, V);
        if Uni then
          -- Rectangle de bords uniformes : remplir uniformement l'interieur
          Fill (Job, V);
          Stack_Top := Stack_Top - 1;

        elsif         Job.Xsup-Job.Xinf < Min_Size
              or else Job.Ysup-Job.Yinf < Min_Size
              or else Stack_Top > Stack'Last - 3 then
          -- Rectangle trop petit pour etre decoupe : evaluation directe
          Interior (Job);
          Stack_Top := Stack_Top - 1;

        else
          -- Couper le rectangle en 4 et le remplacer dans la liste par ces 4
          declare
            Xpiv : constant Gint := (Job.Xsup + Job.Xinf) / 2;
            Ypiv : constant Gint := (Job.Ysup + Job.Yinf) / 2;
            R    : constant Pixrect_Type
                   := (Xinf => Job.Xinf+1, Yinf => Job.Yinf+1,
                       Xsup => Job.Xsup-1, Ysup => Job.Ysup-1);
            HG   : Pixrect_Type renames Stack (Stack_Top + 0);
            HD   : Pixrect_Type renames Stack (Stack_Top + 1);
            BD   : Pixrect_Type renames Stack (Stack_Top + 2);
            BG   : Pixrect_Type renames Stack (Stack_Top + 3);
          begin
            HG := (Xinf=>R.Xinf, Yinf=>R.Yinf, Xsup=>Xpiv,   Ysup=>Ypiv);
            HD := (Xinf=>Xpiv+1, Yinf=>R.Yinf, Xsup=>R.Xsup, Ysup=>Ypiv);
            BD := (Xinf=>Xpiv+1, Yinf=>Ypiv+1, Xsup=>R.Xsup, Ysup=>R.Ysup);
            BG := (Xinf=>R.Xinf, Yinf=>Ypiv+1, Xsup=>Xpiv,   Ysup=>R.Ysup);
            Stack_Top := Stack_Top + 4 - 1;
          end;
        end if;
      end;
    end loop;
    Surface.Flush (S);

    pragma Debug (Put_Line ("Evaluate done"));
  end Evaluate;


  task body Evaluator_Task is
  begin
    loop
      select
        accept Start;
        Evaluate (OWner, Mandelbrot'Access, Owner.Domain);
      or
        accept Done;
      or
        terminate;
      end select;
    end loop;
  end Evaluator_Task;


  package Return_Cb is new Gtk.Handlers.Return_Callback
     (Gadtal_Record, Boolean);

  package Event_Cb is new Gtk.Handlers.User_Return_Callback (
	      Widget_Type => Gtk_Drawing_Area_Record,
	      Return_Type => Boolean,
	      User_Type   => Gadtal_Type);

  package Adj_Cb is new Gtk.Handlers.User_Callback (
            Widget_Type => Gtk_Adjustment_Record,
            User_Type   => Gadtal_Type);

  package Btn_Cb is new Gtk.Handlers.User_Callback (
            Widget_Type => Gtk_Button_Record,
            User_Type   => Gadtal_Type);

  -- delete_event : Destruction de la fenetre principale demandee
  function On_Close (Object : access Gadtal_Record'Class;
                     Params : Gtk.Arguments.Gtk_Args) return Boolean
  is
    pragma Warnings (Off, Object);
    pragma Warnings (Off, Params);
    pragma Debug (Put_Line ("On_Close"));
  begin
    Gtk.Main.Main_Quit;
    return False;		-- Accepte
  end;


  function On_Configure (Object : access Gtk_Drawing_Area_Record'Class;
                         Event  : Gdk_Event;
                         User   : Gadtal_Type) return Boolean
  is
    pragma Warnings (Off, Event);
    pragma Debug (Put_Line ("On_Configure"));
  begin
    -- Le backing store
    Get_Size (Get_Window (Object), User.Width, User.Height);
    if User.Store /= null then Gdk.Pixmap.Unref (User.Store); end if;
    Gdk_New (User.Store, Get_Window(Object), User.Width, User.Height);

    -- Le GC de recopie backing store -> Area
    if User.GC /= null then Unref (User.GC); end if;
    Gdk_New (User.GC, Get_Window (Object));
    return False;
  end;


  function On_Expose (Object : access Gtk_Drawing_Area_Record'Class;
                      Event  : Gdk_Event;
                      User   : Gadtal_Type) return Boolean
  is
    pragma Warnings (Off, Object);
    pragma Warnings (Off, Event);
    pragma Debug (Put_Line ("On_Expose"));
  begin
--  TODO: Select Stop, faire le Draw_Drawable dans la tache si pas de configure
--        entre temps, commander Halt dans Configure, etc
--  User.Evaluator.Start;
--  Evaluate (User, Mandelbrot'Access, User.Domain);
    Draw_Drawable (Drawable => Get_Window (User.Area),
                   Gc       => User.GC,
                   Src      => User.Store,
                   Xsrc     => 0,
                   Ysrc     => 0,
                   Xdest    => 0,
                   Ydest    => 0);

    if Get_Active (User.Alive) then
      if    Get_Value (User.Adj_Depth) <= Get_Lower(User.Adj_Depth) + 1.0
         or Get_Value (User.Adj_Depth) >= Get_Upper(User.Adj_Depth) - 1.0
      then
        User.Going_Up := not User.Going_Up;
      end if;
      if User.Going_Up then
        Set_Value (User.Adj_Depth, Get_Value (User.Adj_Depth) * 1.02);
      else
        Set_Value (User.Adj_Depth, Get_Value (User.Adj_Depth) / 1.02);
      end if;
      Changed (User.Adj_Depth);
      Evaluate (User, Mandelbrot'Access, User.Domain);
      Queue_Draw (User.Area);
    end if;

    return False;
  end;

  function On_Button_Press (Object : access Gtk_Drawing_Area_Record'Class;
                            Event  : Gdk_Event;
                            User   : Gadtal_Type) return Boolean
  is
    pragma Warnings (Off, Object);
    pragma Debug (Put_Line ("On_Button_Press"));
    X      : constant Gint := Gint (Get_X (Event));
    Y      : constant Gint := Gint (Get_Y (Event));
  begin
    User.Rect_Selector.X1 := X;
    User.Rect_Selector.Y1 := Y;
    User.Rect_Selector.X2 := X;
    User.Rect_Selector.Y2 := Y;
    User.Rect_Selector.Selecting := True;
    return False;
  end;

  function On_Button_Release (Object : access Gtk_Drawing_Area_Record'Class;
                              Event  : Gdk_Event;
                              User   : Gadtal_Type) return Boolean
  is
    pragma Warnings (Off, Object);
    pragma Warnings (Off, Event);
    pragma Debug (Put_Line ("On_Button_Release"));
    use Number.Real_CT;
    D  : constant Complex := User.Domain.Max - User.Domain.Min;
    Kx : constant Number.Real := Re(D) / Number.Real (User.Width);
    Ky : constant Number.Real := Im(D) / Number.Real (User.Height);
    function X (I : Gint) return Number.Real is
    begin
      return Re(User.Domain.Min) + Kx * Number.Real(I);
    end;
    function Y (J : Gint) return Number.Real is
    begin
      return Im(User.Domain.Min) + Ky * Number.Real(J);
    end;
  begin
    User.Rect_Selector.Selecting := False;
    if          abs (User.Rect_Selector.X1 - User.Rect_Selector.X2) >= 8
       and then abs (User.Rect_Selector.Y1 - User.Rect_Selector.Y2) >= 8
    then
      User.Domain := (Min => (Re => X(User.Rect_Selector.X1),
                              Im => Y(User.Rect_Selector.Y1)),
                      Max => (Re => X(User.Rect_Selector.X2),
                              Im => Y(User.Rect_Selector.Y2)));

      if Get_Active (User.AutoZoom) then
        Evaluate (User, Mandelbrot'Access, User.Domain);
        Queue_Draw (User.Area);
      end if;
    end if;
    return False;
  end;

  -- Activation de Preferences->Deg
  function On_Motion (Object : access Gtk_Drawing_Area_Record'Class;
                      Event  : Gdk_Event;
                      User   : Gadtal_Type) return Boolean
  is
    pragma Warnings (Off, Object);
    pragma Debug (Put_Line ("On_Move"));
    X      : constant Gint := Gint (Get_X (Event));
    Y      : constant Gint := Gint (Get_Y (Event));
--  Mask   : Gdk.Types.Gdk_Modifier_Type;
--  Result : Gdk_Window;

    procedure Invert is
      X : constant Gint := Gint'Min (User.Rect_Selector.X1,
                                     User.Rect_Selector.X2);
      Y : constant Gint := Gint'Min (User.Rect_Selector.Y1,
                                     User.Rect_Selector.Y2);
      W : constant Gint :=  abs (  User.Rect_Selector.X1
                                 - User.Rect_Selector.X2);
      H : constant Gint :=  abs (  User.Rect_Selector.Y1
                                 - User.Rect_Selector.y2);
    begin
      Draw_Drawable (Drawable => Get_Window (User.Area),
                     Gc       => User.GC,
                     Src      => User.Store,
                     Xsrc     => X,  Xdest    => X,   Width    => W,
                     Ysrc     => Y,  Ydest    => Y,   Height   => H);
    end;

  begin
    if User.Rect_Selector.Selecting then

      Set_Function (User.GC, Invert);
      Invert;
      User.Rect_Selector.X2 := X;
      User.Rect_Selector.Y2 := Y;
      Invert;
      Set_Function (User.GC, Copy);
    end if;
    return False;
  end;

  -- Selection de la profondeur
  procedure On_Depth (Button : access Gtk_Adjustment_Record'Class;
                      User   : Gadtal_Type) is
    pragma Warnings (Off, Button);	-- is not referenced
  begin
    if Get_Active (User.AutoZoom) then
      Evaluate (User, Mandelbrot'Access, User.Domain);
      Queue_Draw (User.Area);
    end if;
  end On_Depth;

  -- Recalculer l'image
  procedure On_Calc (Button : access Gtk_Button_Record'Class;
                     User   : Gadtal_Type) is
    pragma Warnings (Off, Button);	-- is not referenced
  begin
    Evaluate (User, Mandelbrot'Access, User.Domain);
    Queue_Draw (User.Area);
  end On_Calc;


  -- Activation de Preferences->Deg
  type T is access all Gadtal_Record;

  procedure Gtk_New (Gadtal : out Gadtal_Type) is
    VBox         : Gtk_HBox;
    HBox         : Gtk_HBox;
    Spin_Depth   : Gtk_Spin_Button;
    Button       : Gtk_Button;
  begin
    Gadtal := new Gadtal_Record;
    Gtk.Window.Initialize (Gtk_Window (Gadtal), Gtk.Enums.Window_Toplevel);
    Set_Title (Gadtal, "Training to GtkAda");
    Set_Position (Gadtal, Gtk.Enums.Win_Pos_None);
    Set_Modal (Gadtal, False);

    Gadtal.Halt := False;
    Gadtal.Evaluator := new Evaluator_Task (T(Gadtal));
    Gadtal.Width := 0;
    Gadtal.Height := 0;
    Gadtal.Rect_Selector.Selecting := False;
    Gadtal.Domain := Base_Rect;
    Gadtal.Going_Up := False;

    Gtk_New_Hbox (HBox, False, 8);
    Add (Gadtal, HBox);

    Gtk_New (Gadtal.Area);
    Set_Size_Request (Gadtal.Area, 128, 128);
    Event_Cb.Connect (
	  Widget    => Gadtal.Area,
	  Name      => "configure_event",
	  Marsh     => Event_Cb.To_Marshaller (On_Configure'Access),
	  User_Data => Gadtal);
    Event_Cb.Connect (
	  Widget    => Gadtal.Area,
	  Name      => "expose_event",
	  Marsh     => Event_Cb.To_Marshaller (On_Expose'Access),
	  User_Data => Gadtal);
    Event_Cb.Connect (
	  Widget    => Gadtal.Area,
	  Name      => "button_press_event",
	  Marsh     => Event_Cb.To_Marshaller (On_Button_Press'Access),
	  User_Data => Gadtal);
    Event_Cb.Connect (
	  Widget    => Gadtal.Area,
	  Name      => "button_release_event",
	  Marsh     => Event_Cb.To_Marshaller (On_Button_Release'Access),
	  User_Data => Gadtal);
    Event_Cb.Connect (
	  Widget    => Gadtal.Area,
	  Name      => "motion_notify_event",
	  Marsh     => Event_Cb.To_Marshaller (On_Motion'Access),
	  User_Data => Gadtal);
    Set_Events (Gadtal.Area,    Exposure_Mask
			     or Button_Press_Mask
			     or Button_Release_Mask
			     or Scroll_Mask		-- Wheel
			     or Pointer_Motion_Mask);
    Pack_Start (HBox, Gadtal.Area,
                Expand => True, Fill => True, Padding => 0);

    -- Zone des controles
    Gtk_New_Vbox (VBox, False, 8);
    Pack_Start (HBox, VBox,
                Expand => False, Fill => False, Padding => 0);

    -- Selecteur de la profondeur
    Gtk_New (Gadtal.Adj_Depth, Value => 200.0, Lower => 4.0,
             Upper => Gdouble(Value_Type'Last-1),
             Step_Increment => 1.0, Page_Increment => 100.0,
                                    Page_Size      => 100.0);
    Gtk_New (Spin_Depth, Gadtal.Adj_Depth, 1.0, 0);
    Set_Numeric (Spin_Depth, True);
    Set_Update_Policy (Spin_Depth, Update_Always);
    Set_Wrap (Spin_Depth, False);
    Adj_Cb.Connect (
          User_Data => Gadtal,
          Widget    => Gadtal.Adj_Depth,
          Name      => "value_changed",
          Marsh     => Adj_Cb.To_Marshaller (On_Depth'Access));
    Pack_Start (Vbox, Spin_Depth,
                Expand => False, Fill => False, Padding => 0);

    -- Auto-lancement du calcul sur zoom
    Gtk_New (Gadtal.AutoZoom, "Auto calc");
    Pack_Start (Vbox, Gadtal.AutoZoom,
                Expand => False, Fill => False, Padding => 0);

    -- Auto-relancement du calcul avec d'autres parametres
    Gtk_New (Gadtal.Alive, "Alive");
    Set_Mode (Gadtal.Alive, True);
    Pack_Start (Vbox, Gadtal.Alive,
                Expand => False, Fill => False, Padding => 0);

    -- Lancement du calcul
    Gtk_New (Button, "Calcul !");
    Pack_Start (Vbox, Button,
                Expand => False, Fill => False, Padding => 0);
    Btn_Cb.Connect (
       User_Data => Gadtal,
       Widget    => Button,
       Name      => "pressed",
       Marsh     => Btn_Cb.To_Marshaller (On_Calc'Access));

    -- En cas de fermeture de la fenetre
    Return_Cb.Connect (Gadtal, "delete_event", On_Close'Access, False);
    Show_All (Gadtal);
  end;

  procedure Set_Calc (Gadtal : in Gadtal_Type;
                      Rect   : in Number.Rect;
                      Fmap   : in Fmap_Type)
  is
    pragma Warnings (Off, Gadtal);
    pragma Warnings (Off, Rect);
    pragma Warnings (Off, Fmap);
  begin
    null;
  end;

end Gadtal;


